@@ -1,19 +1,7 @@
+// SPDX-License-Identifier: GPL-2.0
 /*
  * Copyright (c) 2000-2002,2005 Silicon Graphics, Inc.
  * All Rights Reserved.
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation.
- *
- * This program is distributed in the hope that it would be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write the Free Software Foundation,
- * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include "libxfs.h"
@@ -35,10 +23,6 @@
 #include "slab.h"
 #include "rmap.h"
 
-#define	rounddown(x, y)	(((x)/(y))*(y))
-
-#define		XR_MAX_SECT_SIZE	(64 * 1024)
-
 /*
  * option tables for getsubopt calls
  */
@@ -155,6 +139,8 @@ err_string(int err_code)
 			_("bad CRC in superblock");
 		err_message[XR_BAD_DIR_SIZE_DATA] =
 			_("inconsistent directory geometry information");
+		err_message[XR_BAD_LOG_GEOMETRY] =
+			_("inconsistent log geometry information");
 		done = 1;
 	}
 
@@ -247,10 +233,16 @@ process_args(int argc, char **argv)
 					if (max_mem_specified)
 						do_abort(
 		_("-o bhash option cannot be used with -m option\n"));
+					if (!val)
+						do_abort(
+		_("-o bhash requires a parameter\n"));
 					libxfs_bhash_size = (int)strtol(val, NULL, 0);
 					bhash_option_used = 1;
 					break;
 				case AG_STRIDE:
+					if (!val)
+						do_abort(
+		_("-o ag_stride requires a parameter\n"));
 					ag_stride = (int)strtol(val, NULL, 0);
 					break;
 				case FORCE_GEO:
@@ -261,6 +253,9 @@ process_args(int argc, char **argv)
 					force_geo = 1;
 					break;
 				case PHASE2_THREADS:
+					if (!val)
+						do_abort(
+		_("-o phase2_threads requires a parameter\n"));
 					phase2_threads = (int)strtol(val, NULL, 0);
 					break;
 				default:
@@ -276,6 +271,9 @@ process_args(int argc, char **argv)
 
 				switch (getsubopt(&p, c_opts, &val)) {
 				case CONVERT_LAZY_COUNT:
+					if (!val)
+						do_abort(
+		_("-c lazycount requires a parameter\n"));
 					lazy_count = (int)strtol(val, NULL, 0);
 					convert_lazy_count = 1;
 					break;
@@ -449,16 +447,15 @@ calc_mkfs(xfs_mount_t *mp)
 	 * ditto the location of the first inode chunks in the fs ('/')
 	 */
 	if (xfs_sb_version_hasdalign(&mp->m_sb) && do_inoalign)  {
-		first_prealloc_ino = XFS_OFFBNO_TO_AGINO(mp, roundup(fino_bno,
-					mp->m_sb.sb_unit), 0);
+		first_prealloc_ino = XFS_AGB_TO_AGINO(mp, roundup(fino_bno,
+					mp->m_sb.sb_unit));
 	} else if (xfs_sb_version_hasalign(&mp->m_sb) &&
 					mp->m_sb.sb_inoalignmt > 1)  {
-		first_prealloc_ino = XFS_OFFBNO_TO_AGINO(mp,
+		first_prealloc_ino = XFS_AGB_TO_AGINO(mp,
 					roundup(fino_bno,
-						mp->m_sb.sb_inoalignmt),
-					0);
+						mp->m_sb.sb_inoalignmt));
 	} else  {
-		first_prealloc_ino = XFS_OFFBNO_TO_AGINO(mp, fino_bno, 0);
+		first_prealloc_ino = XFS_AGB_TO_AGINO(mp, fino_bno);
 	}
 
 	ASSERT(mp->m_ialloc_blks > 0);
@@ -466,7 +463,7 @@ calc_mkfs(xfs_mount_t *mp)
 	if (mp->m_ialloc_blks > 1)
 		last_prealloc_ino = first_prealloc_ino + XFS_INODES_PER_CHUNK;
 	else
-		last_prealloc_ino = XFS_OFFBNO_TO_AGINO(mp, fino_bno + 1, 0);
+		last_prealloc_ino = XFS_AGB_TO_AGINO(mp, fino_bno + 1);
 
 	/*
 	 * now the first 3 inodes in the system
@@ -762,7 +759,7 @@ main(int argc, char **argv)
 
 	chunks_pblock = mp->m_sb.sb_inopblock / XFS_INODES_PER_CHUNK;
 	max_symlink_blocks = libxfs_symlink_blocks(mp, XFS_SYMLINK_MAXLEN);
-	inodes_per_cluster = MAX(mp->m_sb.sb_inopblock,
+	inodes_per_cluster = max(mp->m_sb.sb_inopblock,
 			mp->m_inode_cluster_size >> mp->m_sb.sb_inodelog);
 
 	/*
@@ -860,9 +857,9 @@ main(int argc, char **argv)
 			rlim.rlim_cur = rlim.rlim_max;
 			setrlimit(RLIMIT_AS, &rlim);
 			/* use approximately 80% of rlimit to avoid overrun */
-			max_mem = MIN(max_mem, rlim.rlim_cur / 1280);
+			max_mem = min(max_mem, rlim.rlim_cur / 1280);
 		} else
-			max_mem = MIN(max_mem, (LONG_MAX >> 10) + 1);
+			max_mem = min(max_mem, (LONG_MAX >> 10) + 1);
 
 		if (verbose > 1)
 			do_log(
